﻿using System;
using System.CodeDom.Compiler;
using System.Drawing.Text;
using System.Linq;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using ICSharpCode.TextEditor;
using ICSharpCode.TextEditor.Document;
using StarCoder.Main;
using StarCoder.Main.Sound;
using StarCoder.Main.States.Play;

namespace StarCoder.Editor
{
    public partial class CodeEditor : Form
    {
        private const string REPLACE_MARKER = "<REPLACE_MARKER>";
        protected static readonly Color ERROR_COLOR = Color.OrangeRed;

        private List<EditablePart> editableParts;
        private Game game;
        private ProgrammablePartFactory programmablePartFactory;
        private ITextEditorProperties editorSettings;
        private PrivateFontCollection fonts;

        private class EditablePart
        {
            public TextEditorControl editor;
            public int programmablePartStartIndex;
        }

        private string fullProgrammablePartSource;
        private TextEditorControl currentEditor;

        private Dictionary<string, string> originalToEditablePartReplacementStrings;

        private bool loading = false;

        public CodeEditor()
        {
            InitializeComponent();

            fonts = new PrivateFontCollection();
            fonts.AddFontFile(Path.Combine(Constants.FOLDER_FONTS, "consola.ttf"));

            programmablePartFactory = new ProgrammablePartFactory();

            editableParts = new List<EditablePart>();
            originalToEditablePartReplacementStrings = new Dictionary<string, string>();
            originalToEditablePartReplacementStrings.Add(" && ", " and ");
            originalToEditablePartReplacementStrings.Add(" || ", " or ");
            originalToEditablePartReplacementStrings.Add("!", "not ");

            Reset(false);
            SetEditorSettings();

            SetStartingEditorTabPage(1);

            Load += CodeEditor_Load;
            Closed += CodeEditor_Closed;

            checkBoxSound.Checked = SoundManager.Instance.SoundActive;
        }

        private void SetStartingEditorTabPage(int i)
        {
            tabControlEditors.SelectTab(i);
        }

        private void CodeEditor_Load(object sender, EventArgs e)
        {
            Show();
            Focus();

            game = new Game(panelGame.Handle, programmablePartFactory.CurrentProgrammablePart);

            panelGame.GotFocus += panelGame_GotFocus;
            panelGame.LostFocus += panelGame_LostFocus;

            game.Run();
        }

        private void Reset(bool focusWhenRestarting)
        {
            LoadDocument(Constants.PATH_TO_PROGRAMMABLE_FILE);

            Compile(focusWhenRestarting);
        }

        private void Compile(bool focusWhenRestarting)
        {
            if (loading)
                return;

            if (!IsAnyEditorModified)
            {
                Restart(focusWhenRestarting);
                return;
            }

            foreach (var part in editableParts)
            {
                SetModifiedFlag(part.editor, false);
            }

            string fullClass = fullProgrammablePartSource;

            for (int i = 0; i < editableParts.Count; i++)
            {
                fullClass = fullClass.Replace(REPLACE_MARKER + i, editableParts[i].editor.Text);
            }

            fullClass = ConvertEditedCodeToOriginalCode(fullClass);

            var success = programmablePartFactory.Compile(fullClass);

            if (game != null)
                game.HasCompileErrors = !success;

            MarkErrors();

            if (success)
            {
                Restart(focusWhenRestarting);
            }
        }

        private void MarkErrors()
        {
            bool firstPart = true;
            for (int index = 0; index < editableParts.Count; index++)
            {
                var part = editableParts[index];
                int nextPartStart = int.MaxValue;
                if ((index + 1) < editableParts.Count)
                {
                    nextPartStart = editableParts[index + 1].programmablePartStartIndex;
                }

                var codeEditor = part.editor;
                var programmablePartStartIndex = part.programmablePartStartIndex;

                codeEditor.Document.MarkerStrategy.RemoveAll(marker => marker.Color == ERROR_COLOR);

                if (programmablePartFactory.Errors.Count > 0)
                {
                    var lines = codeEditor.Text.Split('\n');

                    int firstFilledLineIndex;
                    for (firstFilledLineIndex = 0; firstFilledLineIndex < lines.Length; firstFilledLineIndex++)
                    {
                        if (lines[firstFilledLineIndex].Length > 0)
                            break;
                    }

                    foreach (var error in programmablePartFactory.Errors)
                    {
                        if (error.IsWarning)
                            continue;

                        var errorLine = error.Line - programmablePartStartIndex;
                        var errorLineIndex = errorLine - 1;

                        if (errorLineIndex < 0)
                        {
                            if (firstPart)
                            {
                                errorLineIndex = firstFilledLineIndex;
                            }
                            else
                            {
                                continue;
                            }
                        }
                        else if (errorLineIndex < firstFilledLineIndex)
                        {
                            errorLineIndex = firstFilledLineIndex;
                        }
                        else if (errorLineIndex >= lines.Length)
                        {
                            if (errorLineIndex < nextPartStart)
                            {
                                errorLineIndex = lines.Length - 1;
                            }
                            else
                            {
                                continue;
                            }
                        }

                        while ((lines[errorLineIndex].Length == 0) && (errorLineIndex > firstFilledLineIndex))
                        {
                            errorLineIndex--;
                        }

                        int offset = 0;
                        int length = lines[errorLineIndex].Length;
                        for (int i = 0; i < errorLineIndex; i++)
                        {
                            offset += lines[i].Length + 1;
                        }

                        var textMarker = new TextMarker(offset, length, TextMarkerType.SolidBlock, ERROR_COLOR);
                        textMarker.ToolTip = error.ErrorText;
                        codeEditor.Document.MarkerStrategy.AddMarker(textMarker);
                    }
                }

                firstPart = false;
            }

            tabControlEditors.Refresh();
        }

        private void Restart(bool focus)
        {
            if (game == null)
                return;

            game.Restart(programmablePartFactory.CurrentProgrammablePart);

            if (focus)
                panelGame.Focus();
        }

        private void CodeEditor_Closed(object sender, EventArgs e)
        {
            game.Window.Close();
        }

        private void SetEditorSettings()
        {
        }

        private TextEditorControl AddNewTextEditor(string title)
        {
            var tab = new TabPage(title);
            var editor = new TextEditorControl();
            editor.Font = new Font(fonts.Families[0], 9);
            editor.Document.HighlightingStrategy = HighlightingStrategyFactory.CreateHighlightingStrategyForFile("test.cs");
            editor.Dock = DockStyle.Fill;
            editor.IsReadOnly = false;
            editor.Document.DocumentChanged += (sender, e) => DocumentChanged(editor);

            editor.ConvertTabsToSpaces = false;

            editor.ActiveTextAreaControl.Controls.Remove(editor.ActiveTextAreaControl.HScrollBar);
            editor.ActiveTextAreaControl.TextArea.Dock = DockStyle.Left;
            editor.ActiveTextAreaControl.VScrollBar.Dock = DockStyle.Right;
            //editor.ActiveTextAreaControl.Controls.Remove(editor.ActiveTextAreaControl.VScrollBar);
            //editor.ActiveTextAreaControl.TextArea.Dock = DockStyle.Fill;
            

            //panelCode.Controls.Add(editor);

            // When a tab page gets the focus, move the focus to the editor control
            // instead when it gets the Enter (focus) event. I use BeginInvoke 
            // because changing the focus directly in the Enter handler doesn't 
            // work.
            tab.Enter +=
                (sender, e) =>
                    {
                        var page = ((TabPage) sender);
                        page.BeginInvoke(new Action<TabPage>(p => p.Controls[0].Focus()), page);
                        currentEditor = (TextEditorControl) page.Controls[0];
                    };

            tab.Controls.Add(editor);
            tabControlEditors.Controls.Add(tab);

            if (editorSettings == null)
            {
                editorSettings = editor.TextEditorProperties;
            }
            else
            {
                editor.TextEditorProperties = editorSettings;
            }

            if (currentEditor == null)
            {
                currentEditor = editor;
            }

            return editor;
        }

        private void DocumentChanged(TextEditorControl editor)
        {
            SetModifiedFlag(editor, true);

            if (checkBoxAutoCompile.Checked)
            {
                Compile(false);
            }
        }

        private bool IsAnyEditorModified
        {
            get { return editableParts.Any(part => IsModified(part.editor)); }
        }

        private bool IsModified(TextEditorControl editor)
        {
			// TextEditorControl doesn't seem to contain its own 'modified' flag, so 
			// instead we'll treat the "*" on the filename as the modified flag.
			return editor.Parent.Text.EndsWith("*");
		}

        private void SetModifiedFlag(TextEditorControl editor, bool flag)
        {
			if (IsModified(editor) != flag)
			{
				var p = editor.Parent;
				if (IsModified(editor))
					p.Text = p.Text.Substring(0, p.Text.Length - 1);
				else
					p.Text += "*";
			}
		}

        private void LoadDocument(string filename)
        {
            loading = true;

            var fullClassLines = File.ReadAllLines(filename);
            fullProgrammablePartSource = "";

            var startTag = "//START:";
            var endTag = "//END";

            var titles = new List<String>();
            var programmableLineParts = new List<List<String>>();
            var programmablePartStartIndizes = new List<int>();

            var programmable = false;

            var actualLineIndex = 0;

            foreach (var line in fullClassLines)
            {
                if (line.Contains(startTag))
                {
                    programmable = true;
                    fullProgrammablePartSource += REPLACE_MARKER + programmableLineParts.Count + "\n";

                    var titleStartIndex = line.IndexOf(startTag) + startTag.Length;
                    var title = line.Substring(titleStartIndex).Trim();
                    titles.Add(title);

                    programmableLineParts.Add(new List<string>());
                    programmablePartStartIndizes.Add(actualLineIndex);
                }
                else if (line.Contains(endTag))
                {
                    programmable = false;
                }
                else
                {
                    if (programmable)
                    {
                        programmableLineParts[programmableLineParts.Count - 1].Add(line);
                    }
                    else
                    {
                        fullProgrammablePartSource += line + "\n";
                    }

                    actualLineIndex++;
                }
            }

            for (int partIndex = 0; partIndex < programmableLineParts.Count; partIndex++)
            {
                var part = programmableLineParts[partIndex];

                bool allLinesHaveSpace = true;
                while (true)
                {
                    foreach (var line in part)
                    {
                        if ((line.Length > 0) && (!line.StartsWith(" ")))
                        {
                            allLinesHaveSpace = false;
                            break;
                        }
                    }

                    if (!allLinesHaveSpace)
                        break;

                    for (int index = 0; index < part.Count; index++)
                    {
                        var line = part[index];
                        if (line.Length == 0)
                            continue;

                        part[index] = line.Substring(1, line.Length - 1);
                    }
                }

                var programmablePartString = "";
                foreach (var line in part)
                {
                    programmablePartString += line + "\n";
                }

                programmablePartString = programmablePartString.Substring(0, programmablePartString.Length - 1);

                EditablePart editablePart = null;

                if (partIndex >= editableParts.Count)
                {
                    editablePart = new EditablePart();
                    editableParts.Add(editablePart);

                    editablePart.editor = AddNewTextEditor(titles[partIndex]);
                }
                else
                {
                    editablePart = editableParts[partIndex];
                }

                var whiteSpaceToTabRegex = new Regex("([\n\t]) {4}");
                while (whiteSpaceToTabRegex.IsMatch(programmablePartString))
                {
                    programmablePartString = whiteSpaceToTabRegex.Replace(programmablePartString, "$1\t");
                }

                programmablePartString = ConvertOriginalCodeToEditableCode(programmablePartString);

                editablePart.editor.Text = programmablePartString;
                editablePart.programmablePartStartIndex = programmablePartStartIndizes[partIndex];
            }

            loading = false;
        }

        private string ConvertOriginalCodeToEditableCode(string original)
        {
            foreach (var replacement in originalToEditablePartReplacementStrings)
            {
                original = original.Replace(replacement.Key, replacement.Value);
            }
            return original;
        }

        private string ConvertEditedCodeToOriginalCode(string editedCode)
        {
            foreach (var replacement in originalToEditablePartReplacementStrings)
            {
                editedCode = editedCode.Replace(replacement.Value, replacement.Key);
            }
            return editedCode;
        }

        private void panelGame_GotFocus(object sender, EventArgs e)
        {
            Compile(false);
            game.Focus = true;
        }

        private void panelGame_LostFocus(object sender, EventArgs e)
        {
            game.Focus = false;
            Restart(false);
        }

        private void buttonUndo_Click(object sender, EventArgs e)
        {
            currentEditor.Undo();
        }

        private void buttonRedo_Click(object sender, EventArgs e)
        {
            currentEditor.Redo();
        }

        private void buttonRestart_Click(object sender, EventArgs e)
        {
            if (IsAnyEditorModified)
            {
                Compile(true);
            }
            else
            {
                Restart(true);
            }
        }

        private void buttonReset_Click(object sender, EventArgs e)
        {
            Reset(false);
        }

        private void checkBoxSound_CheckedChanged(object sender, EventArgs e)
        {
            SoundManager.Instance.SoundActive = checkBoxSound.Checked;
        }
    }
}
