/*
 * StarCoder by Tobias Wehrum <Tobias.Wehrum@dragonlab.de> is licensed under the
 * Creative Commons Attribution-ShareAlike 3.0 Unported License. To view a copy of this
 * license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a letter to
 * Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;
using SFML.Graphics;
using SFML.Window;
using StarCoder.Main.Sound;
using StarCoder.Main.States;
using StarCoder.Main.States.Play;

namespace StarCoder.Main
{
    public class Game
    {
        public RenderWindow Window { get; protected set; }
        private GameState _currentGameState;
        private GameState nextGameState;

        private Dictionary<String, Texture> cacheTextures;
        private Dictionary<String, Font> cacheFonts;

        private Sprite overlayNoFocus;
        private Text overlayNoFocusText;
        private Sprite overlayCompileErrors;
        private Text overlayCompileErrorsText;
        private Sprite overlayException;
        private Text overlayExceptionText;

        public Random random;
        public float accumulatedFrameTime;

        private bool focus;
        private bool hasCompileErrors;
        private bool hasException;

        public bool PlayFocusSounds { get; set; }
        public bool Updating { get { return Focus && !HasCompileErrors; } }

        protected Font DefaultFont { get; set; }

        public Game(IntPtr handle, IProgrammablePart programmablePart)
        {
            Focus = false;
            HasCompileErrors = false;

            random = new Random();

            var contextSettings = new ContextSettings();
            contextSettings.AntialiasingLevel = 4;

            Window = new RenderWindow(handle, contextSettings);

            /*
            // Create the main window
            if (fullscreen)
            {
                Styles style = Styles.Fullscreen;
                VideoMode desktopMode = VideoMode.DesktopMode;
                uint fullscreenWidth = desktopMode.Width;
                uint fullscreenHeight = desktopMode.Height;

                if (Constants.DEBUG_FULLSCREEN_WINDOW)
                {
                    style = Styles.Default;
                    //fullscreenWidth = 1920;
                    //fullscreenHeight = 768;
                }

                Window = new RenderWindow(new VideoMode(fullscreenWidth, fullscreenHeight), Constants.WINDOW_NAME, style, contextSettings);

                float targetAspectRatio = (float)Constants.TARGET_WIDTH / Constants.TARGET_HEIGHT;
                float currentAspectRatio = (float)fullscreenWidth / fullscreenHeight;

                // Screen to small? Then we will just scale regardless of what would've been the preference.
                if ((fullscreenWidth <= Constants.TARGET_WIDTH) || (fullscreenHeight <= Constants.TARGET_HEIGHT))
                {
                    canUpscale = false;
                    needToDownscale = (fullscreenWidth < Constants.TARGET_WIDTH) || (fullscreenHeight < Constants.TARGET_HEIGHT);
                }

                if (canUpscale || needToDownscale)
                {
                    float viewWidth = Constants.TARGET_WIDTH;
                    float viewHeight = Constants.TARGET_HEIGHT;

                    // Keep aspect ratio
                    if (currentAspectRatio > targetAspectRatio)
                    {
                        viewWidth = viewHeight * currentAspectRatio;
                    }
                    else
                    {
                        viewHeight = viewWidth / currentAspectRatio;
                    }

                    
                    var scaledView = new View(new FloatRect((int)((Constants.TARGET_WIDTH - viewWidth) / 2), (int)((Constants.TARGET_HEIGHT - viewHeight) / 2), viewWidth, viewHeight));

                    if (canUpscale)
                    {
                        viewUpscaled = scaledView;
                    }
                    else
                    {
                        viewDownscaled = scaledView;
                    }
                }

                if (!needToDownscale)
                {
                    viewPixelPerfect = new View(new FloatRect((int) (((int) Constants.TARGET_WIDTH - (int) fullscreenWidth) / 2), (int) (((int) Constants.TARGET_HEIGHT - (int) fullscreenHeight) / 2),
                                                              fullscreenWidth, fullscreenHeight));
                }

                // Center
                //viewRect.Left = 0; viewRect.Top = 0;

                Upscaled = Constants.FULLSCREEN_DEFAULT_UPSCALED;
            }
            else
            {
                // Windowed
                Window = new RenderWindow(new VideoMode(Constants.TARGET_WIDTH, Constants.TARGET_HEIGHT), Constants.WINDOW_NAME, Styles.Close, contextSettings);
            }
            */
            Window.Closed += OnClose;

            Window.SetVerticalSyncEnabled(true);
            Window.SetKeyRepeatEnabled(false);
            Window.SetMouseCursorVisible(false);

            /*
            if (Constants.FRAME_LIMIT > 0)
                Window.SetFramerateLimit(Constants.FRAME_LIMIT);
             */

            accumulatedFrameTime = 0;

            cacheTextures = new Dictionary<string, Texture>();
            cacheFonts = new Dictionary<string, Font>();

            SoundManager.Instance.PreloadSounds();

            focus = false;
            PlayFocusSounds = true;

            DefaultFont = LoadFont("consola.ttf");

            CreateOverlays();

            CurrentGameState = new PlayState(this, programmablePart);
        }

        private void CreateOverlays()
        {
            overlayCompileErrors = CreateOverlaySprite(new Color(255, 0, 0, 150));
            overlayCompileErrorsText = CreateOverlayText("Compile Error");
            
            overlayNoFocus = CreateOverlaySprite(new Color(0, 0, 0, 150));
            overlayNoFocusText = CreateOverlayText("Click here to play.");

            overlayException = CreateOverlaySprite(new Color(255, 0, 0));
            overlayExceptionText = CreateOverlayText("...");
        }

        public Sprite CreateOverlaySprite(Color color)
        {
            var renderTexture = new RenderTexture(Constants.TARGET_WIDTH, Constants.TARGET_HEIGHT);
            renderTexture.Clear(color);

            var sprite = new Sprite(new Texture(renderTexture.Texture));

            renderTexture.Dispose();

            return sprite;
        }

        public Text CreateOverlayText(string message)
        {
            var text = new Text(message, DefaultFont, 32);
            CenterText(text);
            return text;
        }

        public void CenterText(Text text)
        {
            var textBounds = text.GetLocalBounds();
            text.Position = new Vector2f((int)(Constants.TARGET_WIDTH / 2f - textBounds.Width / 2f), (int)(Constants.TARGET_HEIGHT / 2f - textBounds.Height / 2f - 10f));
        }

        public void Run()
        {
            if (Window == null)
                return;

            var clock = new System.Diagnostics.Stopwatch();
            // Start the Game loop
            while (Window.IsOpen())
            {
                Application.DoEvents();

                // Switch states if requested 
                if (nextGameState != null)
                {
                    CurrentGameState = nextGameState;
                    nextGameState = null;
                }

                // Process events
                Window.DispatchEvents();

                float elapsedTimeInS = (float) clock.Elapsed.TotalSeconds;

                if (Updating)
                {
                    accumulatedFrameTime += elapsedTimeInS;
                }
                else
                {
                    accumulatedFrameTime = 0;
                }

                clock.Reset();
                clock.Start();

                if (CurrentGameState != null)
                {
                    try
                    {
                        if (Updating)
                        {
                            while (accumulatedFrameTime > Constants.FRAME_TIME)
                            {
                                accumulatedFrameTime -= Constants.FRAME_TIME;
                                CurrentGameState.Update(elapsedTimeInS);

                                if (nextGameState != null)
                                    break;
                            }
                        }

                        // Clear screen
                        //Window.Clear();

                        CurrentGameState.Draw();
                    }
                    catch (Exception e)
                    {
                        Program.WriteToLogFile("Exception in CurrentGameState.Update or CurrentGameState.Draw:");
                        Program.WriteToLogFile(e.ToString());
                        Program.WriteToLogFile();

                        //MessageBox.Show(e.ToString(), "Crash in Update Loop!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        FoundException("Update Loop");

                        CurrentGameState = null;
                    }
                }

                if (hasException)
                {
                    Window.Draw(overlayException);
                    Window.Draw(overlayExceptionText);
                }
                else if (HasCompileErrors)
                {
                    Window.Draw(overlayCompileErrors);
                    Window.Draw(overlayCompileErrorsText);
                }
                else if (!Focus)
                {
                    Window.Draw(overlayNoFocus);
                    Window.Draw(overlayNoFocusText);
                }

                // Update the window
                Window.Display();
            }
        }

        public void Restart(IProgrammablePart programmablePart)
        {
            hasException = false;
            SwitchToGameState(new PlayState(this, programmablePart));
        }

        /// <summary>
        /// Switch to another Game state. This will be postponed until the current
        /// Game state is rendered and displayed.
        /// </summary>
        /// <param name="gameState">The next GameState</param>
        public void SwitchToGameState(GameState gameState)
        {
            nextGameState = gameState;
        }

        private GameState CurrentGameState
        {
            get { return _currentGameState; }

            set
            {
                if (_currentGameState != null)
                {
                    _currentGameState.TearDown();
                }

                _currentGameState = value;

                if (_currentGameState != null)
                {
                    try
                    {
                        _currentGameState.SetUp();
                    }
                    catch (Exception e)
                    {
                        try
                        {
                            _currentGameState.TearDown();
                            _currentGameState = null;
                        }
                        catch (Exception)
                        {
                        }

                        Program.WriteToLogFile("Exception in CurrentGameState.SetUp:");
                        Program.WriteToLogFile(e.ToString());
                        Program.WriteToLogFile();

                        //MessageBox.Show(e.ToString(), "Crash in Creation!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        FoundException("Creation");
                    }
                }
            }
        }

        protected void OnClose(object sender, EventArgs e)
        {
            // Close the window when OnClose event is received
            //RenderWindow window = (RenderWindow)sender;
            Window.Close();
        }

        /// <summary>
        /// Loads an image and uses a cache on subsequent calls.
        /// </summary>
        /// <param name="path">Path to the image, without "data\\images\\"</param>
        /// <returns>The requested Texture</returns>
        public Texture LoadTexture(String path)
        {
            if (cacheTextures.ContainsKey(path))
            {
                return cacheTextures[path];
            }

            var texture = new Texture(Path.Combine(Constants.FOLDER_IMAGES, path));
            cacheTextures[path] = texture;
            return texture;
        }

        /// <summary>
        /// Loads a font and uses a cache on subsequent calls.
        /// </summary>
        /// <param name="path">Path to the font, without "data\\fonts\\"</param>
        /// <returns>The requested Font</returns>
        public Font LoadFont(String path)
        {
            if (cacheFonts.ContainsKey(path))
            {
                return cacheFonts[path];
            }

            var font = new Font(Path.Combine(Constants.FOLDER_FONTS, path));
            cacheFonts[path] = font;
            return font;
        }

        private void FoundException(string targetSite)
        {
            overlayExceptionText.DisplayedString = "Exception in " + targetSite;
            CenterText(overlayExceptionText);
            hasException = true;
        }

        public bool Focus
        {
            get { return focus; }
            set
            {
                if (focus == value)
                    return;

                focus = value;
                PlaySoundIfStarted();

                if (!focus && PlayFocusSounds)
                {
                    SoundManager.Instance.PlaySound(SoundEvent.LostFocus);
                }
            }
        }

        public bool HasCompileErrors
        {
            get { return hasCompileErrors; }
            set
            {
                if (hasCompileErrors == value)
                    return;

                hasCompileErrors = value;
                PlaySoundIfStarted();
            }
        }

        private void PlaySoundIfStarted()
        {
            if (Updating && PlayFocusSounds)
            {
                SoundManager.Instance.PlaySound(SoundEvent.GameStarted);
            }
        }

        //public float WindowWidth { get { return Window.Size.X; }}
        //public float WindowHeight { get { return Window.Size.Y; } }

        public float Random(float minInclusive, float maxExclusive)
        {
            return minInclusive + (float)random.NextDouble() * (maxExclusive - minInclusive);
        }

        public int Random(int maxExclusive)
        {
            return random.Next(maxExclusive);
        }

        public int Random(int minInclusive, int maxExclusive)
        {
            return random.Next(minInclusive, maxExclusive);
        }
    }
}
