﻿/*
 * StarCoder by Tobias Wehrum <Tobias.Wehrum@dragonlab.de> is licensed under the
 * Creative Commons Attribution-ShareAlike 3.0 Unported License. To view a copy of this
 * license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a letter to
 * Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.Serialization;
using SFML.Audio;
using StarCoder.Main.Helper;

namespace StarCoder.Main.Sound
{
    public class SoundManager
    {
        [Serializable]
        protected class SoundConfig : ISerializable
        {
            private static string FILENAME = "soundConfig";

            private bool _soundActive;

            private SoundConfig(bool soundActive)
            {
                _soundActive = soundActive;
            }

            public SoundConfig(SerializationInfo info, StreamingContext context)
            {
                _soundActive = info.GetBoolean("SoundActive");
            }

            public static SoundConfig Load()
            {
                SoundConfig soundConfig = null;
                if (File.Exists(FILENAME))
                {
                    try
                    {
                        soundConfig = SerializationHelper.DeserializeObject(FILENAME) as SoundConfig;
                    }
                    catch
                    {
                    }
                }
                if (soundConfig == null)
                {
                    soundConfig = new SoundConfig(true);
                }
                return soundConfig;
            }

            public bool Save()
            {
                return SerializationHelper.SerializeObject(FILENAME, this);
            }

            public bool SoundActive
            {
                get { return _soundActive; }
                set
                {
                    if (_soundActive == value)
                        return;

                    _soundActive = value;
                    Save();
                }
            }

            public void GetObjectData(SerializationInfo info, StreamingContext context)
            {
                info.AddValue("SoundActive", _soundActive);
            }
        }

        private static SoundManager _instance;
        public static SoundManager Instance { get { return _instance ?? (_instance = new SoundManager()); } }

        private Dictionary<SoundEvent, string> soundFiles;
        private Dictionary<SoundEvent, SFML.Audio.Sound> sounds;

        private SoundConfig soundConfig;

        private SoundManager()
        {
            soundConfig = SoundConfig.Load();

            soundFiles = new Dictionary<SoundEvent, string>();

            string extension = ".ogg";

            soundFiles[SoundEvent.PlayerDied] = "died" + extension;
            soundFiles[SoundEvent.StarCollected] = "star" + extension;
            soundFiles[SoundEvent.Jump] = "jump" + extension;
            soundFiles[SoundEvent.BumpedHead] = "bumpedHead" + extension;
            //soundFiles[SoundEvent.LostFocus] = "pause" + extension;
            soundFiles[SoundEvent.GameStarted] = "gameStarted" + extension;

            //soundFiles[SoundEvent.Execute] = "Execute" + extension;
            
            /*
            if (Constants.DEBUG_SOUND_DROPBOX) {
                extension = ".wav";
            }
             */

            //soundFiles[SoundEvent.Jump] = "jump" + extension;
        }

        public bool SoundActive
        {
            get { return soundConfig.SoundActive; }
            set { soundConfig.SoundActive = value; }
        }

        public void PreloadSounds()
        {
            sounds = new Dictionary<SoundEvent, SFML.Audio.Sound>();
            foreach (SoundEvent soundEvent in soundFiles.Keys)
            {
                LoadSound(soundEvent);
            }
        }

        private bool LoadSound(SoundEvent soundEvent)
        {
            if (!soundFiles.ContainsKey(soundEvent))
                return false;

            string path = Path.Combine(Constants.FOLDER_SOUNDS, soundFiles[soundEvent]);
            var soundBuffer = new SoundBuffer(path);

            var sound = new SFML.Audio.Sound(soundBuffer);
            sounds[soundEvent] = sound;

            return true;
        }

        public void PlaySound(SoundEvent soundEvent)
        {
            if (!soundConfig.SoundActive)
                return;

            // Reload every time for hot swapping while execution?
            if (Constants.DEBUG_RELOAD_SOUND_EVERY_TIME)
            {
                if (!LoadSound(soundEvent))
                {
                    return;
                }
            }

            if (!sounds.ContainsKey(soundEvent))
            {
                if (!LoadSound(soundEvent))
                {
                    return;
                }
            }

            sounds[soundEvent].Play();
        }
    }
}

