﻿/*
 * StarCoder by Tobias Wehrum <Tobias.Wehrum@dragonlab.de> is licensed under the
 * Creative Commons Attribution-ShareAlike 3.0 Unported License. To view a copy of this
 * license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a letter to
 * Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.
 */

using System;
using System.Collections.Generic;
using SFML.Graphics;
using SFML.Window;

namespace StarCoder.Main.States.Play.Entities
{
    public class Level : PlayStateEntity
    {
        public const int TILE_FREE = -1;
        public const int TILE_WALL = 0;

        private int[][] tiles;

        private Sprite[] sprites;

        public int TileCountX { get; private set; }
        public int TileCountY { get; private set;  }
        public Vector2f PlayerSpawnPosition { get; private set; }

        public Level(PlayState playState, int tileCountX, int tileCountY) : base(playState, new Vector2f())
        {
            TileCountX = tileCountX;
            TileCountY = tileCountY;

            ClearLevel();

            sprites = new Sprite[1];
            sprites[0] = new Sprite(Game.LoadTexture(Constants.ASSET_TILE_WALL));

            Layer = PlayState.Layers.Level;
        }

        private void ClearLevel()
        {
            tiles = new int[TileCountX][];
            for (int x = 0; x < TileCountX; x++)
            {
                tiles[x] = new int[TileCountY];
                for (int y = 0; y < TileCountY; y++)
                {
                    tiles[x][y] = TILE_FREE;
                }
            }
        }

        public void Initialize(string str)
        {
            PlayerSpawnPosition = new Vector2f(Constants.TARGET_WIDTH / 2f, Constants.TARGET_HEIGHT / 2f);

            str = str.ToLower();
            ClearLevel();
            
            for (int i = 0; i < str.Length; i++)
            {
                var c = str[i];
                int x = i % TileCountX;
                int y = i / TileCountX;

                if (y >= TileCountY)
                    return;

                int value = -1;

                switch (c)
                {
                    case 'x':
                        value = TILE_WALL;
                        break;

                    case 's':
                        playState.AddSpikes(new Vector2f((x + 0.5f) * Constants.TILE_SIZE, (y + 1f) * Constants.TILE_SIZE - 1));
                        break;

                    case '*':
                        playState.AddStar(new Vector2f((x + 0.5f) * Constants.TILE_SIZE, (y + 0.5f) * Constants.TILE_SIZE - 1));
                        break;

                    case 'p':
                        PlayerSpawnPosition = new Vector2f((x + 0.5f) * Constants.TILE_SIZE, (y + 1) * Constants.TILE_SIZE - 1);
                        break;
                }

                tiles[x][y] = value;
            }
        }

        public override void Update(float elapsedTime)
        {
        }

        public Vector2f MoveRect(FloatRect rect, Vector2f delta)
        {
            var originalPosition = new Vector2f(rect.Left, rect.Top);
            var direction = new Vector2i(Math.Sign(delta.X), Math.Sign(delta.Y));

            while (Math.Abs(delta.X) != 0)
            {
                float oldValue = rect.Left;
                float shift = direction.X;

                if (Math.Abs(shift) > Math.Abs(delta.X))
                {
                    shift = delta.X;
                    delta.X = 0;
                }
                else
                {
                    delta.X -= shift;
                }

                rect.Left += shift;

                if (CollisionWithSolid(rect))
                {
                    rect.Left = oldValue;
                    break;
                }
            }
            
            while (Math.Abs(delta.Y) != 0)
            {
                float oldValue = rect.Top;
                float shift = direction.Y;

                if (Math.Abs(shift) > Math.Abs(delta.Y))
                {
                    shift = delta.Y;
                    delta.Y = 0;
                }
                else
                {
                    delta.Y -= shift;
                }

                rect.Top += shift;

                if (CollisionWithSolid(rect))
                {
                    rect.Top = oldValue;
                    break;
                }
            }

            return new Vector2f(rect.Left - originalPosition.X, rect.Top - originalPosition.Y);
        }

        public HashSet<int> GetCollisions(FloatRect rect)
        {
            var tileRect = ToTileRect(rect);

            var collisions = new HashSet<int>();
            DoForTileRect(tileRect, delegate(int x, int y, int tile)
                                        {
                                            if (tile != TILE_FREE)
                                            {
                                                collisions.Add(tile);
                                            }
                                            return true;
                                        });
            return collisions;
        }

        private bool CollisionWithSolid(FloatRect rect)
        {
            bool collision = false;

            var tileRect = ToTileRect(rect);
            DoForTileRect(tileRect, delegate(int x, int y, int tile)
                                        {
                                            if (tile == TILE_WALL)
                                            {
                                                collision = true;
                                                return false;
                                            }
                                            return true;
                                        });

            return collision;
        }

        private delegate bool TileRectAction(int x, int y, int tile);

        private void DoForTileRect(IntRect tileRect, TileRectAction action)
        {
            for (int x = tileRect.Left; x <= tileRect.Left + tileRect.Width; x++)
            {
                for (int y = tileRect.Top; y <= tileRect.Top + tileRect.Height; y++)
                {
                    if (!action(x, y, tiles[x][y]))
                        return;
                }
            }
        }

        public IntRect ToTileRect(FloatRect worldRect)
        {
            int startX = Math.Max((int) (worldRect.Left / Constants.TILE_SIZE), 0);
            int startY = Math.Max((int) (worldRect.Top / Constants.TILE_SIZE), 0);
            int endX = Math.Min((int) Math.Floor((worldRect.Left + worldRect.Width) / Constants.TILE_SIZE), TileCountX - 1);
            int endY = Math.Min((int) Math.Floor((worldRect.Top + worldRect.Height) / Constants.TILE_SIZE), TileCountY - 1);

            return new IntRect(startX, startY, endX - startX, endY - startY);
        }

        public override void Draw(RenderTarget renderTarget)
        {
            for (int x = 0; x < TileCountX; x++)
            {
                for (int y = 0; y < TileCountY; y++)
                {
                    var tile = tiles[x][y];
                    if (tile == TILE_FREE)
                        continue;

                    var sprite = sprites[tile];
                    sprite.Position = new Vector2f(x * Constants.TILE_SIZE, y * Constants.TILE_SIZE);
                    renderTarget.Draw(sprite);
                }
            }
        }
    }
}
