﻿/*
 * StarCoder by Tobias Wehrum <Tobias.Wehrum@dragonlab.de> is licensed under the
 * Creative Commons Attribution-ShareAlike 3.0 Unported License. To view a copy of this
 * license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a letter to
 * Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.
 */

using System;
using SFML.Graphics;
using SFML.Window;
using StarCoder.Main.Helper;
using StarCoder.Main.Sound;

namespace StarCoder.Main.States.Play.Entities
{
    public class Player : PlayStateEntity
    {
        public bool Standing { get; private set; }

        public Sprite Sprite;
        private Vector2f gravityDirection;
        private Vector2f velocity;

        public Vector2f Movement { get; set; }
        public Vector2f Velocity
        {
            get { return velocity; }
            set { velocity = value; }
        }

        public uint Width { get; private set; }
        public uint Height { get; private set; }

        public FloatRect Rect { get { return new FloatRect(Position.X - Width / 2f, Position.Y - Height / 2f, Width, Height); } }

        public Player(PlayState playState, Vector2f position)
            : base(playState, position)
        {
            var texture = Game.LoadTexture(Constants.ASSET_PLAYER);

            Width = texture.Size.X;
            Height = texture.Size.Y;

            position.Y -= Height / 2f;

            Sprite = new Sprite(texture);
            Sprite.Origin = new Vector2f(Width / 2f, Height / 2f);

            Layer = PlayState.Layers.Player;
        }

        public override void Update(float elapsedTime)
        {
        }

        public override bool Move(float dx, float dy)
        {
            var delta = playState.Level.MoveRect(Rect, new Vector2f(dx, dy));

            Standing = false;
            if (gravityDirection.Y > 0)
            {
                if ((dy > 0.0001f) && (delta.Y < (dy / 2f)))
                {
                    Standing = true;
                }
            }
            else if (gravityDirection.Y < 0)
            {
                if ((dy < -0.0001f) && (delta.Y > (dy / 2f)))
                {
                    Standing = true;
                }
            }

            if (gravityDirection.X > 0)
            {
                if ((dx > 0.0001f) && (delta.X < (dx / 2f)))
                {
                    Standing = true;
                }
            }
            else if (gravityDirection.X < 0)
            {
                if ((dx < -0.0001f) && (delta.X > (dx / 2f)))
                {
                    Standing = true;
                }
            }

            if (gravityDirection.Y != 0)
            {
                if ((Math.Abs(dy) / 2f) > Math.Abs(delta.Y))
                {
                    if ((Velocity.Y != 0) && !Standing)
                    {
                        SoundManager.Instance.PlaySound(SoundEvent.BumpedHead);
                    }
                    velocity.Y = 0;
                }
            }

            if (gravityDirection.X != 0)
            {
                if ((Math.Abs(dx) / 2f) > Math.Abs(delta.X))
                {
                    if ((Velocity.X != 0) && !Standing)
                    {
                        SoundManager.Instance.PlaySound(SoundEvent.BumpedHead);
                    }
                    velocity.X = 0;
                }
            }

            var result = base.Move(delta.X, delta.Y);

            WrapMovement(delta, Rect);
            if (position.Y == (Constants.TARGET_HEIGHT - 1))
            {
                Standing = true;
            }

            return result;
        }

        public override void Draw(RenderTarget renderTarget)
        {
            Sprite.Position = position;
            renderTarget.Draw(Sprite);
        }

        public void Jump(float jumpImpulse)
        {
            Standing = false;
            velocity.X = GravityDirection.X * jumpImpulse;
            velocity.Y = GravityDirection.Y * jumpImpulse;

            SoundManager.Instance.PlaySound(SoundEvent.Jump);
        }

        public void ResetVelocityX()
        {
            velocity.X = 0;
        }

        public void ResetVelocityY()
        {
            velocity.Y = 0;
        }

        public Vector2f GravityDirection
        {
            get { return gravityDirection; }
            set
            {
                if ((gravityDirection.X == value.X) && (gravityDirection.Y == value.Y))
                    return;

                gravityDirection = value;

                float angle = (float) (Math.Atan2(gravityDirection.Y, gravityDirection.X) - Math.PI / 2);
                Sprite.Rotation = angle * MathHelper.RAD_TO_DEG;

                float originalWidth = Sprite.Texture.Size.X;
                float originalHeight = Sprite.Texture.Size.Y;

                Width = (uint)Math.Round(Math.Abs(originalWidth * Math.Cos(angle)) + Math.Abs(originalHeight * Math.Sin(angle)));
                Height = (uint)Math.Round(Math.Abs(originalWidth * Math.Sin(angle)) + Math.Abs(originalHeight * Math.Cos(angle)));
            }
        }
    }
}
