﻿/*
 * StarCoder by Tobias Wehrum <Tobias.Wehrum@dragonlab.de> is licensed under the
 * Creative Commons Attribution-ShareAlike 3.0 Unported License. To view a copy of this
 * license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a letter to
 * Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.
 */

using System.Collections.Generic;
using System.Linq;
using SFML.Graphics;
using SFML.Window;
using StarCoder.Main.Entities;
using StarCoder.Main.Sound;
using StarCoder.Main.States.Play.Entities;

namespace StarCoder.Main.States.Play
{
    public class PlayState : GameState
    {
        private readonly IProgrammablePart programmablePart;

        public class Layers
        {
            public const int Level = 2;
            public const int Entities = 4;
            public const int Player = 5;
            public const int Gui = 10;
        }

        protected EntityList<Entity> EntityList { get; set; }

        public Player Player { get; protected set; }
        public Level Level { get; protected set; }
        public List<Star> Stars { get; protected set; }
        public List<Spike> Spikes { get; protected set; }

        public bool PlayerWon { get; set; }

        private Sprite overlayWon;
        private Text overlayWonText;
        private Sprite background;

        public PlayState(Game game, IProgrammablePart programmablePart) : base(game)
        {
            this.programmablePart = programmablePart;
            EntityList = new EntityList<Entity>();
            Spikes = new List<Spike>();
            Stars = new List<Star>();

            overlayWon = Game.CreateOverlaySprite(new Color(255, 255, 0, 150));
            overlayWonText = Game.CreateOverlayText("You did it!");
            overlayWonText.Color = Color.Black;
        }

        public void AddEntity(Entity entity)
        {
            EntityList.Add(entity);
        }

        public void RemoveEntity(Entity entity)
        {
            EntityList.Remove(entity);
        }

        public override void SetUp()
        {
            background = new Sprite(Game.LoadTexture(Constants.ASSET_BACKGROUND));

            Player = new Player(this, new Vector2f(Constants.TARGET_WIDTH / 2f, Constants.TARGET_HEIGHT / 2f));
            AddEntity(Player);

            Level = new Level(this, (int) (Constants.TARGET_WIDTH / Constants.TILE_SIZE), (int) (Constants.TARGET_HEIGHT / Constants.TILE_SIZE));
            AddEntity(Level);

            var coordinateSystem = new Sprite(Game.LoadTexture(Constants.ASSET_COORDINATE_SYSTEM));
            coordinateSystem.Position = new Vector2f(0, Constants.TARGET_HEIGHT - coordinateSystem.Texture.Size.Y);
            AddEntity(new DrawableEntity(coordinateSystem, Layers.Gui));

            Window.KeyPressed += WindowOnKeyPressed;
            Window.MouseButtonPressed += WindowOnMouseButtonPressed;

            programmablePart.SetUp(this);
        }

        public override void TearDown()
        {
            Window.KeyPressed -= WindowOnKeyPressed;
            Window.MouseButtonPressed -= WindowOnMouseButtonPressed;

            programmablePart.TearDown();
        }

        public override void Update(float elapsedTime)
        {
            if (PlayerWon)
            {
                return;
            }

            programmablePart.Update(elapsedTime);
            EntityList.Update(elapsedTime);
        }

        private void WindowOnKeyPressed(object sender, KeyEventArgs args)
        {
            if (PlayerWon)
            {
                SoundManager.Instance.PlaySound(SoundEvent.GameStarted);
                Restart();
            }
        }
        
        private void WindowOnMouseButtonPressed(object sender, MouseButtonEventArgs mouseButtonEventArgs)
        {
            if (PlayerWon)
            {
                SoundManager.Instance.PlaySound(SoundEvent.GameStarted);
                Restart();
            }
        }

        public void Restart()
        {
            Game.SwitchToGameState(new PlayState(Game, programmablePart));
        }

        public override void Draw()
        {
            Window.Draw(background);

            EntityList.Draw(Window);

            if (PlayerWon)
            {
                Window.Draw(overlayWon);
                Window.Draw(overlayWonText);
            }
        }

        public bool CollidesWithSpikes(FloatRect rect)
        {
            return Spikes.Any(spike => spike.Rect.Intersects(rect));
        }

        public void AddSpikes(Vector2f position)
        {
            var newSpikes = new Spike(this, position);
            Spikes.Add(newSpikes);
            AddEntity(newSpikes);
        }

        public bool CollidesWithStar(FloatRect rect)
        {
            return Stars.Any(star => star.Rect.Intersects(rect));
        }

        public void AddStar(Vector2f position)
        {
            if (PlayerWon)
                return;

            var newStar = new Star(this, position);
            Stars.Add(newStar);
            AddEntity(newStar);
        }

        public void RemoveStar()
        {
            foreach (var star in Stars)
            {
                RemoveEntity(star);
            }
            Stars.Clear();
        }
    }
}
