﻿/*
 * StarCoder by Tobias Wehrum <Tobias.Wehrum@dragonlab.de> is licensed under the
 * Creative Commons Attribution-ShareAlike 3.0 Unported License. To view a copy of this
 * license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a letter to
 * Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.
 */

using System.Collections.Generic;
using SFML.Window;
using StarCoder.Main.Sound;
using StarCoder.Main.States.Play.Entities;
using StarCoder.Main.Helper;
using Key = SFML.Window.Keyboard.Key;

namespace StarCoder.Main.States.Play
{
    public class ProgrammablePart : IProgrammablePart
    {
        private PlayState playState;
        private Player player;
        private Vector2f actualGravity;
        private double gravity;
        private HashSet<Key> keysPressedThisFrame;

        public void SetUp(PlayState playState)
        {
            this.playState = playState;
            player = playState.Player;

            string level = "";
            gravity = 0;

            {
                //START: Creation
                // This is just called once when the game starts

                // Initialize level
                level = "xxxxxxxxxx" +
                        "x        x" +
                        "x        x" +
                        "x        x" +
                        "x        x" +
                        "x P     *x" +
                        "x        x" +
                        "x        x" +
                        "x        x" +
                        "xx      xx" +
                        "xxxSSSSxxx" +
                        "xxxxxxxxxx";

                // Set the gravity
                gravity = -7;
                //END
            }

            playState.Level.Initialize(level);
            player.Set(playState.Level.PlayerSpawnPosition);

            keysPressedThisFrame = new HashSet<Key>();
            playState.Window.KeyPressed += KeyPressed;
        }

        public void TearDown()
        {
            playState.Window.KeyPressed -= KeyPressed;
        }

        public void Update(float elapsedTime)
        {
            actualGravity.X = 0;
            actualGravity.Y = 0;

            {
                //START: Update Loop
                // All of this is called 60 times per second

                // If Left is held down, move left by 2 pixels with MovePlayer(x, y)
                if (IsKeyDown(Key.Left))
                {
                    MovePlayer(-2, 0);
                }

                // If Right is held down, move right by 2 pixels with MovePlayer(x, y)
                if (IsKeyDown(Key.Right))
                {
                    MovePlayer(2, 0);
                }

                // If Space was just pressed and the player is standing, jump
                if (WasKeyPressedJustNow(Key.Space) && IsPlayerStanding)
                {
                    JumpWithSpeed(7);
                }

                // Apply gravity to the player
                ApplyGravityY(gravity);

                // If the player touches a star, he collects that star
                if (PlayerTouchesStar)
                {
                    WinGame();
                }

                // If the player touches spikes, kill and respawn him
                if (PlayerTouchesSpikes)
                {
                    KillAndRespawnPlayer();
                }
                //END
            }

            /*
            if (IsKeyDown(Key.Up)) MovePlayer(0, 2);
            if (IsKeyDown(Key.Down)) MovePlayer(0, -2);
            if (WasKeyPressedJustNow(Key.W)) gravity = -gravity;
             */

            player.Velocity += new Vector2f(actualGravity.X * 0.03f, actualGravity.Y * 0.03f);
            player.GravityDirection = new Vector2f(actualGravity.X, actualGravity.Y).GetNormalized();

            player.Move(player.Velocity.X + player.Movement.X, player.Velocity.Y + player.Movement.Y);
            player.Movement = new Vector2f();
            player.Update(elapsedTime);

            keysPressedThisFrame.Clear();
        }

        private bool IsKeyDown(Key key)
        {
            return Keyboard.IsKeyPressed(key);
        }

        private bool WasKeyPressedJustNow(Key key)
        {
            return keysPressedThisFrame.Contains(key);
        }

        private void MovePlayer(double x, double y)
        {
            player.Movement += new Vector2f((float)x, (float)-y);

            if (x != 0)
            {
                player.ResetVelocityX();
            }
            if (y != 0)
            {
                player.ResetVelocityY();
            }
        }

        private void MoveStar(double x, double y)
        {
            MoveStars(x, y);
        }

        private void MoveStars(double x, double y)
        {
            foreach (var star in playState.Stars)
            {
                star.Move((float) x, (float) -y);
            }
        }

        private void MoveSpikes(double x, double y)
        {
            foreach (var spikes in playState.Spikes)
            {
                spikes.Move((float) x, (float) -y);
            }
        }

        private void JumpWithSpeed(double jumpSpeed)
        {
            player.Jump((float) -jumpSpeed);
        }

        private void ApplyGravityX(double x)
        {
            actualGravity.X += (float)x;
        }

        private void ApplyGravityY(double y)
        {
            actualGravity.Y += (float) -y;
        }

        private void KillAndRespawnPlayer()
        {
            if (playState.PlayerWon)
                return;

            SoundManager.Instance.PlaySound(SoundEvent.PlayerDied);
            playState.Restart();
        }

        private void WinGame()
        {
            playState.RemoveStar();

            playState.PlayerWon = true;
            SoundManager.Instance.PlaySound(SoundEvent.StarCollected);
        }

        protected bool IsPlayerStanding
        {
            get { return player.Standing; }
        }

        protected bool IsPlayerJumping
        {
            get { return !IsPlayerStanding; }
        }

        public bool PlayerTouchesSpikes
        {
            get { return playState.CollidesWithSpikes(player.Rect); }
        }

        public bool PlayerTouchesStar
        {
            get { return playState.CollidesWithStar(player.Rect); }
        }

        private void KeyPressed(object sender, KeyEventArgs e)
        {
            keysPressedThisFrame.Add(e.Code);
        }
    }
}
